from distclassipy.classifier import DistanceMetricClassifier

import numpy as np

import pytest

from sklearn.utils.estimator_checks import check_estimator


# Test initialization of the classifier with specific parameters
def test_init():
    clf = DistanceMetricClassifier(metric="euclidean", scale=True)
    assert clf.metric == "euclidean"
    assert clf.scale is True


def test_sklearn_compatibility():
    check_estimator(DistanceMetricClassifier())


# Test fitting the classifier to a dataset
def test_fit():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier()
    clf.fit(X, y)
    assert clf.is_fitted_ is True
    assert clf.n_features_in_ == 2


# Test making predictions with the classifier
def test_predict():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier()
    clf.fit(X, y)
    predictions = clf.predict(X)
    assert len(predictions) == len(y)


# Test fitting and predicting without scaling std
def test_predict_without_stdscale():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier(scale=False)
    clf.fit(X, y)
    predictions = clf.predict(X)
    assert clf.scale is False
    assert len(predictions) == len(y)


# Test calculating confidence of predictions
def test_calculate_confidence():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier(calculate_kde=True)
    clf.fit(X, y)
    clf.predict_and_analyse(X)
    confidence = clf.calculate_confidence()
    assert confidence.shape == (3, len(np.unique(y)))


# Test using different distance metrics - from scipy
def test_metric_scipy():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier(metric="cityblock")
    clf.fit(X, y)
    assert clf.metric == "cityblock"


# Test using different distance metrics - from distclassipy
def test_metric_dcpy():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier(metric="soergel")
    clf.fit(X, y)
    assert clf.metric == "soergel"


# Test using custom defined metric
def test_metric_custom():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values

    def metric_euc(u, v):
        return np.sqrt(np.sum((u - v) ** 2))

    clf = DistanceMetricClassifier(metric=metric_euc)
    clf.fit(X, y)
    assert callable(clf.metric)


# Test using invalid metric
def test_metric_invalid():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values

    with pytest.raises(ValueError):
        clf = DistanceMetricClassifier(metric="chaini")
        clf.fit(X, y)


# Test setting central statistical method to median
def test_central_stat_median():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier(central_stat="median", dispersion_stat="iqr")
    clf.fit(X, y)
    assert clf.central_stat == "median"


# Test setting central statistical method to mean
def test_central_stat_mean():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier(central_stat="mean")
    clf.fit(X, y)
    assert clf.central_stat == "mean"


# Test KDE calculation functionality
def test_kde_calculation():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier(calculate_kde=True)
    clf.fit(X, y)
    clf.predict_and_analyse(X)
    assert hasattr(clf, "kde_dict_")


# Test 1D distance calculation functionality
def test_1d_distance_calculation():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier(calculate_1d_dist=True)
    clf.fit(X, y)
    clf.predict_and_analyse(X)
    assert hasattr(clf, "conf_cl_")


# Test prediction error when the classifier is not fitted
def test_predict_without_fit():
    clf = DistanceMetricClassifier()
    with pytest.raises(ValueError):
        clf.predict(np.array([[1, 2]]))


# Test confidence calculation error when analysis is not performed
def test_calculate_confidence_without_analysis():
    clf = DistanceMetricClassifier()
    with pytest.raises(ValueError):
        clf.calculate_confidence()


# Test different methods of confidence calculation
def test_confidence_calculation_methods():
    X = np.array([[1, 2], [3, 4], [5, 6]])  # Sample feature set
    y = np.array([0, 1, 0])  # Sample target values
    clf = DistanceMetricClassifier(calculate_kde=True, calculate_1d_dist=True)
    clf.fit(X, y)
    clf.predict_and_analyse(X)
    distance_confidence = clf.calculate_confidence(method="distance_inverse")
    assert distance_confidence.shape == (3, len(np.unique(y)))
    kde_confidence = clf.calculate_confidence(method="kde_likelihood")
    assert kde_confidence.shape == (3, len(np.unique(y)))
    one_d_confidence = clf.calculate_confidence(method="1d_distance_inverse")
    assert one_d_confidence.shape == (3, len(np.unique(y)))



# Note: A large portion of this code has been inspired
# from https://github.com/aziele/statistical-distance/blob/f49212e/test.py

from distclassipy.distances import Distance

import numpy as np

import pytest

distance = Distance()

# Test case 1: non-zero values in u and v.
uv1 = [np.array([0.33, 0.21, 0.46]), np.array([0.32, 0.50, 0.18])]

# Test case 2: zero value in v.
uv2 = [np.array([0.41, 0.23, 0.36]), np.array([0.30, 0.70, 0])]

# Test case 3: zero value in u.
uv3 = [np.array([0.33, 0.67, 0]), np.array([0.50, 0.25, 0.25])]

# Test case 4: zero value in u and v.
uv4 = [np.array([0.45, 0.55, 0]), np.array([0.68, 0.32, 0])]

# Test case 5: u and v are identical.
uv5 = [
    np.array([0.20, 0.05, 0.40, 0.30, 0.05]),
    np.array([0.20, 0.05, 0.40, 0.30, 0.05]),
]


@pytest.fixture
def vectors():
    return [uv1, uv2, uv3, uv4, uv5]


def func_test(func, vectors, correct_values):
    for i, (u, v) in enumerate(vectors):
        np.testing.assert_almost_equal(func(u, v), correct_values[i], decimal=6)


def test_euclidean(vectors):
    correct_values = [0.403237, 0.602163, 0.517494, 0.325269, 0]
    func_test(distance.euclidean, vectors, correct_values)


def test_braycurtis(vectors):
    correct_values = [0.29, 0.47, 0.42, 0.23, 0]
    func_test(distance.braycurtis, vectors, correct_values)


def test_canberra(vectors):
    correct_values = [0.861335, 1.660306, 1.661341, 0.467908, 0]
    func_test(distance.canberra, vectors, correct_values)


def test_cityblock(vectors):
    correct_values = [0.58, 0.94, 0.84, 0.46, 0]
    func_test(distance.cityblock, vectors, correct_values)


def test_chebyshev(vectors):
    correct_values = [0.29, 0.47, 0.42, 0.23, 0]
    func_test(distance.chebyshev, vectors, correct_values)


def test_clark(vectors):
    correct_values = [0.598728, 1.131109, 1.118196, 0.333645, 0]
    func_test(distance.clark, vectors, correct_values)


def test_correlation(vectors):
    correct_values = [1.995478, 1.755929, 1.008617, 0.254193, 0]
    func_test(distance.correlation, vectors, correct_values)


def test_cosine(vectors):
    correct_values = [0.216689, 0.370206, 0.272996, 0.097486, 0]
    func_test(distance.cosine, vectors, correct_values)


def test_hellinger(vectors):
    correct_values = [0.5029972, 0.996069, 0.859140, 0.330477, 0]
    func_test(distance.hellinger, vectors, correct_values)


def test_jaccard(vectors):
    correct_values = [0.356579, 0.560779, 0.446110, 0.179993, 0]
    func_test(distance.jaccard, vectors, correct_values)


def test_lorentzian(vectors):
    correct_values = [0.511453, 0.797107, 0.730804, 0.414028, 0]
    func_test(distance.lorentzian, vectors, correct_values)


def test_marylandbridge(vectors):
    correct_values = [0.216404, 0.350152, 0.258621, 0.096073, 0]
    func_test(distance.marylandbridge, vectors, correct_values)


def test_meehl(vectors):
    correct_values = [0.1629, 0.3989, 0.3545, 0.2645, 0]
    func_test(distance.meehl, vectors, correct_values)


def test_motyka(vectors):
    correct_values = [0.645, 0.7349999, 0.71, 0.615, 0.5]
    func_test(distance.motyka, vectors, correct_values)


def test_soergel(vectors):
    correct_values = [0.449612, 0.639456, 0.591549, 0.373984, 0]
    func_test(distance.soergel, vectors, correct_values)


def test_wave_hedges(vectors):
    correct_values = [1.218999, 1.939721, 1.966866, 0.756417, 0]
    func_test(distance.wave_hedges, vectors, correct_values)


def test_kulczynski(vectors):
    correct_values = [0.816901, 1.773585, 1.448276, 0.597403, 0]
    func_test(distance.kulczynski, vectors, correct_values)


def test_add_chisq(vectors):
    correct_values = [1.175282, 1.345852, 1.114259, 0.456844, 0]
    func_test(distance.add_chisq, vectors, correct_values)



"""A module for using distance metrics for classification.

Classes:
    DistanceMetricClassifier - A classifier that uses a specified distance metric for
                               classification.
    Distance - A class that provides various distance metrics for use in classification.


Copyright (C) 2024  Siddharth Chaini
-----
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .classifier import DistanceMetricClassifier  # noqa
from .distances import Distance  # noqa

__version__ = "0.1.6a0"



"""A module providing a variety of distance metrics.

This module includes implementations of various distance metrics, including both
common and less common measures. It allows for the calculation of distances between
data points in a vectorized manner using numpy arrays.
A part of this code is based on the work of Andrzej Zielezinski, originally retrieved
on 20 November 2022 from
https://github.com/aziele/statistical-distances/blob/04412b3155c59fc7238b3d8ecf6f3723ac5befff/distance.py, # noqa
which was released via the GNU General Public License v3.0.

It was originally modified by Siddharth Chaini on 27 November 2022.

Notes
-----
    Modifications by Siddharth Chaini include the addition of the following distance
    measures:
        1. Meehl distance
        2. Sorensen distance
        3. Ruzicka distance
        4. Inner product distance
        5. Harmonic mean distance
        6. Fidelity
        7. Minimimum Symmetric Chi Squared
        8. Probabilistic Symmetric Chi Squared

    In addition, the following code was added to all functions for array conversion:
        u,v = np.asarray(u), np.asarray(v)

Copyright (C) 2024  Siddharth Chaini
-----
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import warnings

import numpy as np

import scipy


class Distance:
    """A class to calculate various distance metrics between vectors.

    This class provides methods to compute different types of distances between
    two vectors, such as Euclidean, Manhattan, Canberra, and other statistical
    distances. Each method takes two vectors as input and returns the calculated
    distance. The class can handle both numpy arrays and lists, converting them
    internally to numpy arrays for computation.

    Attributes
    ----------
    epsilon : float, optional
        A small value to avoid division by zero errors in certain distance
        calculations. Default is the machine precision for float data type.

    Methods
    -------
    acc(u, v)
        Returns the average of Cityblock/Manhattan and Chebyshev distances.
    add_chisq(u, v)
        Returns the Additive Symmetric Chi-square distance.
    (Other methods are not listed here for brevity)

    Examples
    --------
    >>> dist = Distance()
    >>> u = [1, 2, 3]
    >>> v = [4, 5, 6]
    >>> print(dist.acc(u, v))
    5.0
    """

    def __init__(self, epsilon=None):
        """Initialize the Distance class with an optional epsilon value.

        Parameters
        ----------
        - epsilon: A small value to avoid division by zero errors.
        """
        self.epsilon = np.finfo(float).eps if not epsilon else epsilon

    def euclidean(self, u, v, w=None):
        """Calculate the Euclidean distance between two vectors.

        The Euclidean distance is the "ordinary" straight-line distance between two
        points in Euclidean space.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Euclidean distance between the two vectors.

        References
        ----------
            1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4), 300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        return scipy.spatial.distance.euclidean(u, v, w)

    def braycurtis(self, u, v, w=None):
        """Calculate the Bray-Curtis distance between two vectors.

        The Bray-Curtis distance is a measure of dissimilarity between two non-negative
        vectors, often used in ecology to measure the compositional dissimilarity
        between two sites based on counts of species at both sites. It is closely
        related to the Sørensen distance and is also known as Bray-Curtis
        dissimilarity.

        Notes
        -----
            When used for comparing two probability density functions (pdfs),
            the Bray-Curtis distance equals the Cityblock distance divided by 2.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Bray-Curtis distance between the two vectors.

        References
        ----------
            1. Bray JR, Curtis JT (1957) An ordination of the upland forest of
               southern Wisconsin. Ecological Monographs, 27, 325-349.
            2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4), 300-307.
            3. https://en.wikipedia.org/wiki/Bray–Curtis_dissimilarity
        """
        u, v = np.asarray(u), np.asarray(v)
        return scipy.spatial.distance.braycurtis(u, v, w)

    def canberra(self, u, v, w=None):
        """Calculate the Canberra distance between two vectors.

        The Canberra distance is a weighted version of the Manhattan distance, used
        in numerical analysis.

        Notes
        -----
            When `u[i]` and `v[i]` are 0 for given i, then the fraction 0/0 = 0
            is used in the calculation.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Canberra distance between the two vectors.

        References
        ----------
            1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4), 300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        return scipy.spatial.distance.canberra(u, v, w)

    def cityblock(self, u, v, w=None):
        """Calculate the Cityblock (Manhattan) distance between two vectors.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Cityblock distance between the two vectors.

        References
        ----------
            1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4):300-307.

        Synonyms:
            City block distance
            Manhattan distance
            Rectilinear distance
            Taxicab norm

        Notes
        -----
            Cityblock distance between two probability density functions
            (pdfs) equals:
            1. Non-intersection distance multiplied by 2.
            2. Gower distance multiplied by vector length.
            3. Bray-Curtis distance multiplied by 2.
            4. Google distance multiplied by 2.
        """
        u, v = np.asarray(u), np.asarray(v)
        return scipy.spatial.distance.cityblock(u, v, w)

    def chebyshev(self, u, v, w=None):
        """Calculate the Chebyshev distance between two vectors.

        The Chebyshev distance is a metric defined on a vector space where the distance
        between two vectors
        is the greatest of their differences along any coordinate dimension.

        Synonyms:
            Chessboard distance
            King-move metric
            Maximum value distance
            Minimax approximation

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Chebyshev distance between the two vectors.

        References
        ----------
            1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4), 300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        return scipy.spatial.distance.chebyshev(u, v, w)

    def correlation(self, u, v, w=None, centered=True):
        """Calculate the Pearson correlation distance between two vectors.

        Returns a distance value between 0 and 2.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Pearson correlation distance between the two vectors.
        """
        u, v = np.asarray(u), np.asarray(v)
        if len(u) < 2 or len(v) < 2:
            warnings.warn(
                "Pearson correlation requires vectors of length at least 2.",
                RuntimeWarning,
            )
            d = 0
        else:
            d = scipy.spatial.distance.correlation(u, v, w, centered)
            if np.isnan(d) and (
                np.allclose(u - np.mean(u), 0) or np.allclose(v - np.mean(v), 0)
            ):
                warnings.warn(
                    "One of the vectors is constant; correlation is set to 0",
                    RuntimeWarning,
                )
                d = 0
        return d

    def cosine(self, u, v, w=None):
        """Calculate the cosine distance between two vectors.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The cosine distance between the two vectors.

        References
        ----------
            1. SciPy.
        """
        u, v = np.asarray(u), np.asarray(v)
        return scipy.spatial.distance.cosine(u, v, w)

    def clark(self, u, v):
        """Calculate the Clark distance between two vectors.

        The Clark distance equals the square root of half of the divergence.

        Notes
        -----
            When `u[i]` and `v[i]` are 0 for given i, then the fraction 0/0 = 0
            is used in the calculation.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Clark distance between the two vectors.

        References
        ----------
            1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4), 300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        with np.errstate(divide="ignore", invalid="ignore"):
            return np.sqrt(np.nansum(np.power(np.abs(u - v) / (u + v), 2)))

    def hellinger(self, u, v):
        """Calculate the Hellinger distance between two vectors.

        The Hellinger distance is a measure of similarity between two probability
        distributions.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Hellinger distance between the two vectors.

        Notes
        -----
            This implementation produces values two times larger than values
            obtained by Hellinger distance described in Wikipedia and also
            in https://gist.github.com/larsmans/3116927.

        References
        ----------
           1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
              Measures between Probability Density Functions. International
              Journal of Mathematical Models and Methods in Applied Sciences.
              1(4), 300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        return np.sqrt(2 * np.sum((np.sqrt(u) - np.sqrt(v)) ** 2))

    def jaccard(self, u, v):
        """Calculate the Jaccard distance between two vectors.

        The Jaccard distance measures dissimilarity between sample sets.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Jaccard distance between the two vectors.

        References
        ----------
           1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
              Measures between Probability Density Functions. International
              Journal of Mathematical Models and Methods in Applied Sciences.
              1(4), 300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        uv = np.dot(u, v)
        return 1 - (uv / (np.dot(u, u) + np.dot(v, v) - uv))

    def lorentzian(self, u, v):
        """Calculate the Lorentzian distance between two vectors.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Lorentzian distance between the two vectors.

        References
        ----------
            1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4):300-307.

        Notes
        -----
            One (1) is added to guarantee the non-negativity property and to
            eschew the log of zero.
        """
        u, v = np.asarray(u), np.asarray(v)
        return np.sum(np.log(np.abs(u - v) + 1))

    def marylandbridge(self, u, v):
        """Calculate the Maryland Bridge distance between two vectors.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Maryland Bridge distance between the two vectors.

        References
        ----------
            1. Deza M, Deza E (2009) Encyclopedia of Distances.
               Springer-Verlag Berlin Heidelberg. 1-590.
        """
        u, v = np.asarray(u), np.asarray(v)
        uvdot = np.dot(u, v)
        return 1 - (uvdot / np.dot(u, u) + uvdot / np.dot(v, v)) / 2

    def meehl(self, u, v):
        """Calculate the Meehl distance between two vectors.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Meehl distance between the two vectors.

        Notes
        -----
            Added by SC.

        References
        ----------
            1. Deza M. and Deza E. (2013) Encyclopedia of Distances.
               Berlin, Heidelberg: Springer Berlin Heidelberg.
               https://doi.org/10.1007/978-3-642-30958-8.
        """
        u, v = np.asarray(u), np.asarray(v)

        xi = u[:-1]
        yi = v[:-1]
        xiplus1 = np.roll(u, 1)[:-1]
        yiplus1 = np.roll(v, 1)[:-1]

        with np.errstate(divide="ignore", invalid="ignore"):
            return np.nansum((xi - yi - xiplus1 + yiplus1) ** 2)

    def motyka(self, u, v):
        """Calculate the Motyka distance between two vectors.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Motyka distance between the two vectors.

        Notes
        -----
            The distance between identical vectors is not equal to 0 but 0.5.

        References
        ----------
            1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4), 300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        return np.sum(np.maximum(u, v)) / np.sum(u + v)

    def soergel(self, u, v):
        """Calculate the Soergel distance between two vectors.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Soergel distance between the two vectors.

        Notes
        -----
            Equals Tanimoto distance.

        References
        ----------
            1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4), 300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        return np.sum(np.abs(u - v)) / np.sum(np.maximum(u, v))

    def wave_hedges(self, u, v):
        """Calculate the Wave Hedges distance between two vectors.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Wave Hedges distance between the two vectors.

        References
        ----------
            1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4), 300-307
        """
        u, v = np.asarray(u), np.asarray(v)
        with np.errstate(divide="ignore", invalid="ignore"):
            u_v = abs(u - v)
            uvmax = np.maximum(u, v)
            return np.sum(np.where(((u_v != 0) & (uvmax != 0)), u_v / uvmax, 0))

    def kulczynski(self, u, v):
        """Calculate the Kulczynski distance between two vectors.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Kulczynski distance between the two vectors.

        References
        ----------
            1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions. International
               Journal of Mathematical Models and Methods in Applied Sciences.
               1(4):300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        return np.sum(np.abs(u - v)) / np.sum(np.minimum(u, v))

    def add_chisq(self, u, v):
        """Compute the Additive Symmetric Chi-square distance between two vectors.

        The Additive Symmetric Chi-square distance is a measure that
        can be used to compare two vectors. This function calculates it based
        on the input vectors u and v.

        Parameters
        ----------
        - u, v: Input vectors between which the distance is to be calculated.

        Returns
        -------
        - The Additive Symmetric Chi-square distance between the two vectors.

        References
        ----------
            1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity
               Measures between Probability Density Functions.
               International Journal of Mathematical Models and Methods in
               Applied Sciences.
               vol. 1(4), pp. 300-307.
        """
        u, v = np.asarray(u), np.asarray(v)
        uvmult = u * v
        with np.errstate(divide="ignore", invalid="ignore"):
            return np.sum(np.where(uvmult != 0, ((u - v) ** 2 * (u + v)) / uvmult, 0))


# NOTE: THE FOLLOWING DISTANCES HAVE NOT BEEN TESTED EXTENSIVELY AND ARE
# CURRENTLY IN ALPHA AND SO HAVE BEEN COMMENTED

#     def acc(self, u, v):
#         """Calculate the average of Cityblock and Chebyshev distance.

#         This function computes the ACC distance, also known as the
#         Average distance, between two vectors u and v. It is the average of the
#         Cityblock (or Manhattan) and Chebyshev distances.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The ACC distance between the two vectors.

#         References
#         ----------
#             1. Krause EF (2012) Taxicab Geometry An Adventure in Non-Euclidean
#                Geometry. Dover Publications.
#             2. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                vol. 1(4), pp. 300-307.
#         """
#         return (self.cityblock(u, v) + self.chebyshev(u, v)) / 2

#     # def bhattacharyya(self, u, v):
#     #     """
#     #     Calculate the Bhattacharyya distance between two vectors.

#     #     Returns a distance value between 0 and 1.

#     #     Parameters
#     #     ----------
#     #     - u, v: Input vectors between which the distance is to be calculated.

#     #     Returns
#     #     -------
#     #     - The Bhattacharyya distance between the two vectors.

#     #     References
#     #     ----------
#     #         1. Bhattacharyya A (1947) On a measure of divergence between two
#     #            statistical populations defined by probability distributions,
#     #            Bull. Calcutta Math. Soc., 35, 99–109.
#     #         2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#     #            Measures between Probability Density Functions. International
#     #            Journal of Mathematical Models and Methods in Applied Sciences.
#     #            1(4), 300-307.
#     #         3. https://en.wikipedia.org/wiki/Bhattacharyya_distance
#     #     """
#     #     u, v = np.asarray(u), np.asarray(v)
#     #     return -np.log(np.sum(np.sqrt(u * v)))

#     def chebyshev_min(self, u, v):
#         """Calculate the minimum value distance between two vectors.

#         This measure represents a custom approach by Zielezinski to distance
#         measurement, focusing on the minimum absolute difference.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The minimum value distance between the two vectors.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return np.amin(np.abs(u - v))

#     def czekanowski(self, u, v):
#         """Calculate the Czekanowski distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Czekanowski distance between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return np.sum(np.abs(u - v)) / np.sum(u + v)

#     def dice(self, u, v):
#         """Calculate the Dice dissimilarity between two vectors.

#         Synonyms:
#             Sorensen distance

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Dice dissimilarity between the two vectors.

#         References
#         ----------
#             1. Dice LR (1945) Measures of the amount of ecologic association
#                between species. Ecology. 26, 297-302.
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         u_v = u - v
#         return np.dot(u_v, u_v) / (np.dot(u, u) + np.dot(v, v))

#     def divergence(self, u, v):
#         """Calculate the divergence between two vectors.

#         Divergence equals squared Clark distance multiplied by 2.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The divergence between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         with np.errstate(invalid="ignore"):
#             return 2 * np.nansum(np.power(u - v, 2) / np.power(u + v, 2))

#     # def fidelity(self, u, v):
#     #     """
#     #     Calculate the fidelity distance between two vectors.

#     #     The fidelity distance measures the similarity between two probability
#     #     distributions.

#     #     Parameters
#     #     ----------
#     #     - u, v: Input vectors between which the distance is to be calculated.

#     #     Returns
#     #     -------
#     #     - The fidelity distance between the two vectors.

#     #     Notes
#     #     -----
#     #         Added by SC.
#     #     """
#     #     u, v = np.asarray(u), np.asarray(v)
#     #     return 1 - (np.sum(np.sqrt(u * v)))

#     def google(self, u, v):
#         """Calculate the Normalized Google Distance (NGD) between two vectors.

#         NGD is a measure of similarity derived from the number of hits returned by the
#         Google search engine for a given set of keywords.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Normalized Google Distance between the two vectors.

#         Notes
#         -----
#             When used for comparing two probability density functions (pdfs),
#             Google distance equals half of Cityblock distance.

#         References
#         ----------
#             1. Lee & Rashid (2008) Information Technology, ITSim 2008.
#                doi:10.1109/ITSIM.2008.4631601.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         x = float(np.sum(u))
#         y = float(np.sum(v))
#         summin = float(np.sum(np.minimum(u, v)))
#         return (max([x, y]) - summin) / ((x + y) - min([x, y]))

#     def gower(self, u, v):
#         """Calculate the Gower distance between two vectors.

#         The Gower distance equals the Cityblock distance divided by the vector length.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Gower distance between the two vectors.

#         References
#         ----------
#             1. Gower JC. (1971) General Coefficient of Similarity
#                and Some of Its Properties, Biometrics 27, 857-874.
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return np.sum(np.abs(u - v)) / u.size

#     # NEEDS CHECKING
#     # def harmonicmean(self, u, v):
#     #     """
#     #     Harmonic mean distance.
#     #     Notes:
#     #         Added by SC.
#     #     """
#     #     u,v = np.asarray(u), np.asarray(v)
#     #     return 1 - 2.*np.sum(u*v/(u+v))

#     # def inner(self, u, v):
#     #     """
#     #     Calculate the inner product distance between two vectors.

#     #     The inner product distance is a measure of similarity between two vectors,
#     #     based on their inner product.

#     #     Parameters
#     #     ----------
#     #     - u, v: Input vectors between which the distance is to be calculated.

#     #     Returns
#     #     -------
#     #     - The inner product distance between the two vectors.

#     #     Notes
#     #     -----
#     #         Added by SC.
#     #     """
#     #     u, v = np.asarray(u), np.asarray(v)
#     #     return 1 - np.dot(u, v)

#     def jeffreys(self, u, v):
#         """Calculate the Jeffreys divergence between two vectors.

#         The Jeffreys divergence is a symmetric version of the Kullback-Leibler
#         divergence.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the divergence is to be calculated.

#         Returns
#         -------
#         - The Jeffreys divergence between the two vectors.

#         References
#         ----------
#             1. Jeffreys H (1946) An Invariant Form for the Prior Probability
#                in Estimation Problems. Proc.Roy.Soc.Lon., Ser. A 186, 453-461.
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         # Add epsilon to zeros in vectors to avoid division
#         # by 0 and/or log of 0. Alternatively, zeros in the
#         # vectors could be ignored or masked (see below).
#         # u = ma.masked_where(u == 0, u)
#         # v = ma.masked_where(v == 0, u)
#         u = np.where(u == 0, self.epsilon, u)
#         v = np.where(v == 0, self.epsilon, v)
#         return np.sum((u - v) * np.log(u / v))

#     def jensenshannon_divergence(self, u, v):
#         """Calculate the Jensen-Shannon divergence between two vectors.

#         The Jensen-Shannon divergence is a symmetric and finite measure of similarity
#         between two probability distributions.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the divergence is to be calculated.

#         Returns
#         -------
#         - The Jensen-Shannon divergence between the two vectors.

#         References
#         ----------
#             1. Lin J. (1991) Divergence measures based on the Shannon entropy.
#                IEEE Transactions on Information Theory, 37(1):145–151.
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         Comments:
#             Equals Jensen difference in Sung-Hyuk (2007):
#             u = np.where(u==0, self.epsilon, u)
#             v = np.where(v==0, self.epsilon, v)
#             el1 = (u * np.log(u) + v * np.log(v)) / 2
#             el2 = (u + v)/2
#             el3 = np.log(el2)
#             return np.sum(el1 - el2 * el3)
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         u = np.where(u == 0, self.epsilon, u)
#         v = np.where(v == 0, self.epsilon, v)
#         dl = u * np.log(2 * u / (u + v))
#         dr = v * np.log(2 * v / (u + v))
#         return (np.sum(dl) + np.sum(dr)) / 2

#     def jensen_difference(self, u, v):
#         """Calculate the Jensen difference between two vectors.

#         The Jensen difference is considered similar to the Jensen-Shannon divergence.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Jensen difference between the two vectors.

#         Notes
#         -----
#             1. Equals half of Topsøe distance
#             2. Equals squared jensenshannon_distance.


#         References
#         ----------
#             1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         u = np.where(u == 0, self.epsilon, u)
#         v = np.where(v == 0, self.epsilon, v)
#         el1 = (u * np.log(u) + v * np.log(v)) / 2
#         el2 = (u + v) / 2
#         return np.sum(el1 - el2 * np.log(el2))

#     def k_divergence(self, u, v):
#         """Calculate the K divergence between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the divergence is to be calculated.

#         Returns
#         -------
#         - The K divergence between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         u = np.where(u == 0, self.epsilon, u)
#         v = np.where(v == 0, self.epsilon, v)
#         return np.sum(u * np.log(2 * u / (u + v)))

#     def kl_divergence(self, u, v):
#         """Calculate the Kullback-Leibler divergence between two vectors.

#         The Kullback-Leibler divergence measures the difference between two
#         probability distributions.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the divergence is to be calculated.

#         Returns
#         -------
#         - The Kullback-Leibler divergence between the two vectors.

#         References
#         ----------
#             1. Kullback S, Leibler RA (1951) On information and sufficiency.
#                Ann. Math. Statist. 22:79–86
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4):300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         u = np.where(u == 0, self.epsilon, u)
#         v = np.where(v == 0, self.epsilon, v)
#         return np.sum(u * np.log(u / v))

#     def kumarjohnson(self, u, v):
#         """Calculate the Kumar-Johnson distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Kumar-Johnson distance between the two vectors.

#         References
#         ----------
#             1. Kumar P, Johnson A. (2005) On a symmetric divergence measure
#                and information inequalities, Journal of Inequalities in pure
#                and applied Mathematics. 6(3).
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4):300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         uvmult = u * v
#         with np.errstate(divide="ignore", invalid="ignore"):
#             numer = np.power(u**2 - v**2, 2)
#             denom = 2 * np.power(uvmult, 3 / 2)
#             return np.sum(np.where(uvmult != 0, numer / denom, 0))

#     def matusita(self, u, v):
#         """Calculate the Matusita distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Matusita distance between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4):300-307.

#         Notes
#         -----
#             Equals square root of Squared-chord distance.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return np.sqrt(np.sum((np.sqrt(u) - np.sqrt(v)) ** 2))

#     def max_symmetric_chisq(self, u, v):
#         """Calculate the maximum symmetric chi-square distance.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The maximum symmetric chi-square distance between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4):300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return max(self.neyman_chisq(u, v), self.pearson_chisq(u, v))

#     def min_symmetric_chisq(self, u, v):
#         """Calculate the minimum symmetric chi-square distance.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The minimum symmetric chi-square distance between the two vectors.

#         Notes
#         -----
#             Added by SC.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return min(self.neyman_chisq(u, v), self.pearson_chisq(u, v))

#     def minkowski(self, u, v, p=2):
#         """Calculate the Minkowski distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.
#         - p: The order of the norm of the difference.

#         Returns
#         -------
#         - The Minkowski distance between the two vectors.

#         Notes
#         -----
#             When p goes to infinite, the Chebyshev distance is derived.

#         References
#         ----------
#             1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4):300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return np.linalg.norm(u - v, ord=p)

#     def neyman_chisq(self, u, v):
#         """Calculate the Neyman chi-square distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Neyman chi-square distance between the two vectors.

#         References
#         ----------
#             1. Neyman J (1949) Contributions to the theory of the chi^2 test.
#                In Proceedings of the First Berkley Symposium on Mathematical
#                Statistics and Probability.
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         with np.errstate(divide="ignore", invalid="ignore"):
#             return np.sum(np.where(u != 0, (u - v) ** 2 / u, 0))

#     # def nonintersection(self, u, v):
#     #     """
#     #     Calculate the Nonintersection distance between two vectors.

#     #     Parameters
#     #     ----------
#     #     - u, v: Input vectors between which the distance is to be calculated.

#     #     Returns
#     #     -------
#     #     - The Nonintersection distance between the two vectors.

#     #     References
#     #     ----------
#     #         1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#     #            Measures between Probability Density Functions. International
#     #            Journal of Mathematical Models and Methods in Applied Sciences.
#     #            1(4), 300-307.

#     #     Notes
#     #     -----
#     #         When used for comparing two probability density functions (pdfs),
#     #         Nonintersection distance equals half of Cityblock distance.
#     #     """
#     #     u, v = np.asarray(u), np.asarray(v)
#     #     return 1 - np.sum(np.minimum(u, v))

#     def pearson_chisq(self, u, v):
#         """Calculate the Pearson chi-square divergence between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the divergence is to be calculated.

#         Returns
#         -------
#         - The Pearson chi-square divergence between the two vectors.

#         References
#         ----------
#             1. Pearson K. (1900) On the Criterion that a given system of
#                deviations from the probable in the case of correlated system
#                of variables is such that it can be reasonable supposed to have
#                arisen from random sampling, Phil. Mag. 50, 157-172.
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.

#         Notes
#         -----
#             Pearson chi-square divergence is asymmetric.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         with np.errstate(divide="ignore", invalid="ignore"):
#             return np.sum(np.where(v != 0, (u - v) ** 2 / v, 0))

#     def penroseshape(self, u, v):
#         """Calculate the Penrose shape distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Penrose shape distance between the two vectors.

#         References
#         ----------
#             1. Deza M, Deza E (2009) Encyclopedia of Distances.
#                Springer-Verlag Berlin Heidelberg. 1-590.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         umu = np.mean(u)
#         vmu = np.mean(v)
#         return np.sqrt(np.sum(((u - umu) - (v - vmu)) ** 2))

#     def prob_chisq(self, u, v):
#         """Calculate the Probabilistic chi-square distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Probabilistic chi-square distance between the two vectors.

#         Notes
#         -----
#             Added by SC.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         uvsum = u + v
#         with np.errstate(divide="ignore", invalid="ignore"):
#             return 2 * np.sum(np.where(uvsum != 0, (u - v) ** 2 / uvsum, 0))

#     def ruzicka(self, u, v):
#         """Calculate the Ruzicka distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Ruzicka distance between the two vectors.

#         Notes
#         -----
#             Added by SC.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         den = np.sum(np.maximum(u, v))

#         return 1 - np.sum(np.minimum(u, v)) / den

#     def sorensen(self, u, v):
#         """Calculate the Sorensen distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Sorensen distance between the two vectors.

#         Notes
#         -----
#             The Sorensen distance equals the Manhattan distance divided by the sum of
#             the two vectors.

#             Added by SC.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return np.sum(np.abs(u - v)) / np.sum(u + v)

#     def squared_chisq(self, u, v):
#         """Calculate the Squared chi-square distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Squared chi-square distance between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         uvsum = u + v
#         with np.errstate(divide="ignore", invalid="ignore"):
#             return np.sum(np.where(uvsum != 0, (u - v) ** 2 / uvsum, 0))

#     def squaredchord(self, u, v):
#         """Calculate the Squared-chord distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Squared-chord distance between the two vectors.

#         References
#         ----------
#             1. Gavin DG et al. (2003) A statistical approach to evaluating
#                distance metrics and analog assignments for pollen records.
#                Quaternary Research 60:356–367.
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.

#         Notes
#         -----
#             Equals to squared Matusita distance.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return np.sum((np.sqrt(u) - np.sqrt(v)) ** 2)

#     def squared_euclidean(self, u, v):
#         """Calculate the Squared Euclidean distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Squared Euclidean distance between the two vectors.

#         References
#         ----------
#             1. Gavin DG et al. (2003) A statistical approach to evaluating
#                distance metrics and analog assignments for pollen records.
#                Quaternary Research 60:356–367.
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.

#         Notes
#         -----
#             Equals to squared Euclidean distance.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         return np.dot((u - v), (u - v))

#     def taneja(self, u, v):
#         """Calculate the Taneja distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Taneja distance between the two vectors.

#         References
#         ----------
#             1. Taneja IJ. (1995), New Developments in Generalized Information
#                Measures, Chapter in: Advances in Imaging and Electron Physics,
#                Ed. P.W. Hawkes, 91, 37-135.
#             2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         u = np.where(u == 0, self.epsilon, u)
#         v = np.where(v == 0, self.epsilon, v)
#         uvsum = u + v
#         return np.sum((uvsum / 2) * np.log(uvsum / (2 * np.sqrt(u * v))))

#     def tanimoto(self, u, v):
#         """Calculate the Tanimoto distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Tanimoto distance between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.

#         Notes
#         -----
#             Equals Soergel distance.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         # return np.sum(abs(u-v)) / np.sum(np.maximum(u, v))
#         usum = np.sum(u)
#         vsum = np.sum(v)
#         minsum = np.sum(np.minimum(u, v))
#         return (usum + vsum - 2 * minsum) / (usum + vsum - minsum)

#     def topsoe(self, u, v):
#         """Calculate the Topsøe distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Topsøe distance between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.

#         Notes
#         -----
#             Equals two times Jensen-Shannon divergence.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         u = np.where(u == 0, self.epsilon, u)
#         v = np.where(v == 0, self.epsilon, v)
#         dl = u * np.log(2 * u / (u + v))
#         dr = v * np.log(2 * v / (u + v))
#         return np.sum(dl + dr)

#     def vicis_symmetric_chisq(self, u, v):
#         """Calculate the Vicis Symmetric chi-square distance.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Vicis Symmetric chi-square distance between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         with np.errstate(divide="ignore", invalid="ignore"):
#             u_v = (u - v) ** 2
#             uvmin = np.minimum(u, v) ** 2
#             return np.sum(np.where(uvmin != 0, u_v / uvmin, 0))

#     def vicis_wave_hedges(self, u, v):
#         """Calculate the Vicis-Wave Hedges distance between two vectors.

#         Parameters
#         ----------
#         - u, v: Input vectors between which the distance is to be calculated.

#         Returns
#         -------
#         - The Vicis-Wave Hedges distance between the two vectors.

#         References
#         ----------
#             1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity
#                Measures between Probability Density Functions. International
#                Journal of Mathematical Models and Methods in Applied Sciences.
#                1(4), 300-307.
#         """
#         u, v = np.asarray(u), np.asarray(v)
#         with np.errstate(divide="ignore", invalid="ignore"):
#             u_v = abs(u - v)
#             uvmin = np.minimum(u, v)
#             return np.sum(np.where(uvmin != 0, u_v / uvmin, 0))



"""A module containing the distance metric classifier.

This module contains the DistanceMetricClassifier introduced by Chaini et al. (2024)
in "Light Curve Classification with DistClassiPy: a new distance-based classifier"

Copyright (C) 2024  Siddharth Chaini
-----
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import warnings
from typing import Callable

import numpy as np

import pandas as pd

import scipy

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.neighbors import KernelDensity
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted

from .distances import Distance

# Hardcoded source packages to check for distance metrics.
METRIC_SOURCES_ = {
    "scipy.spatial.distance": scipy.spatial.distance,
    "distances.Distance": Distance(),
}


class DistanceMetricClassifier(BaseEstimator, ClassifierMixin):
    """A distance-based classifier that supports different distance metrics.

    The distance metric classifier determines the similarity between features in a
    dataset by leveraging the use of different distance metrics to. A specified
    distance metric is used to compute the distance between a given object and a
    centroid for every training class in the feature space. The classifier supports
    the use of different statistical measures for constructing the centroid and scaling
    the computed distance. Additionally, the distance metric classifier also
    optionally provides an estimate of the confidence of the classifier's predictions.

    Parameters
    ----------
    metric : str or callable, default="euclidean"
        The distance metric to use for calculating the distance between features.
    scale : bool, default=True
        Whether to scale the distance between the test object and the centroid for a
        class in the feature space. If True, the data will be scaled based on the
        specified dispersion statistic.
    central_stat : {"mean", "median"}, default="median"
        The statistic used to calculate the central tendency of the data to construct
        the feature-space centroid. Supported statistics are "mean" and "median".
    dispersion_stat : {"std", "iqr"}, default="std"
        The statistic used to calculate the dispersion of the data for scaling the
        distance. Supported  statistics are "std" for standard deviation and "iqr"
        for inter-quartile range.

        .. versionadded:: 0.1.0

    calculate_kde : bool, default=False
        Whether to calculate a kernel density estimate based confidence parameter.
        .. deprecated:: 0.2.0
            This parameter will be removed in a future version and only the
            distance confidence parameter will be available.
    calculate_1d_dist : bool, default=False
        Whether to calculate the 1-dimensional distance based confidence parameter.
        .. deprecated:: 0.2.0
            This parameter will be removed in a future version and only the
            distance confidence parameter will be available.
        Whether to calculate the 1-dimensional distance based confidence parameter.

    Attributes
    ----------
    metric : str or callable
        The distance metric used for classification.
    scale : bool
        Indicates whether the data is scaled.
    central_stat : str
        The statistic used for calculating central tendency.
    dispersion_stat : str
        The statistic used for calculating dispersion.
    calculate_kde : bool
        Indicates whether a kernel density estimate is calculated.
        .. deprecated:: 0.2.0
            This parameter will be removed in a future version.
    calculate_1d_dist : bool
        Indicates whether 1-dimensional distances are calculated.
        .. deprecated:: 0.2.0
            This parameter will be removed in a future version.

    See Also
    --------
    scipy.spatial.dist : Other distance metrics provided in SciPy
    distclassipy.Distance : Distance metrics included with DistClassiPy

    Notes
    -----
    If using distance metrics supported by SciPy, it is desirable to pass a string,
    which allows SciPy to use an optimized C version of the code instead of the slower
    Python version.

    References
    ----------
    .. [1] "Light Curve Classification with DistClassiPy: a new distance-based
            classifier"

    Examples
    --------
    >>> import distclassipy as dcpy
    >>> from sklearn.datasets import make_classification
    >>> X, y = make_classification(n_samples=1000, n_features=4,
    ...                            n_informative=2, n_redundant=0,
    ...                            random_state=0, shuffle=False)
    >>> clf = dcpy.DistanceMetricClassifier(metric="canberra")
    >>> clf.fit(X, y)
    DistanceMetricClassifier(...)
    >>> print(clf.predict([[0, 0, 0, 0]]))
    [0]
    """

    def __init__(
        self,
        metric: str | Callable = "euclidean",
        scale: bool = True,
        central_stat: str = "median",
        dispersion_stat: str = "std",
        calculate_kde: bool = True,  # deprecated in 0.2.0
        calculate_1d_dist: bool = True,  # deprecated in 0.2.0
    ):
        """Initialize the classifier with specified parameters."""
        self.metric = metric
        self.scale = scale
        self.central_stat = central_stat
        self.dispersion_stat = dispersion_stat
        if calculate_kde:
            warnings.warn(
                "calculate_kde is deprecated and will be removed in version 0.2.0",
                DeprecationWarning,
            )
        self.calculate_kde = calculate_kde

        if calculate_1d_dist:
            warnings.warn(
                "calculate_1d_dist is deprecated and will be removed in version 0.2.0",
                DeprecationWarning,
            )
        self.calculate_1d_dist = calculate_1d_dist

    def initialize_metric_function(self):
        """Set the metric function based on the provided metric.

        If the metric is a string, the function will look for a corresponding
        function in scipy.spatial.distance or distances.Distance. If the metric
        is a function, it will be used directly.
        """
        if callable(self.metric):
            self.metric_fn_ = self.metric
            self.metric_arg_ = self.metric

        elif isinstance(self.metric, str):
            metric_str_lowercase = self.metric.lower()
            metric_found = False
            for package_str, source in METRIC_SOURCES_.items():

                # Don't use scipy for jaccard as their implementation only works with
                # booleans - use custom jaccard instead
                if (
                    package_str == "scipy.spatial.distance"
                    and metric_str_lowercase == "jaccard"
                ):
                    continue

                if hasattr(source, metric_str_lowercase):
                    self.metric_fn_ = getattr(source, metric_str_lowercase)
                    metric_found = True

                    # Use the string as an argument if it belongs to scipy as it is
                    # optimized
                    self.metric_arg_ = (
                        self.metric
                        if package_str == "scipy.spatial.distance"
                        else self.metric_fn_
                    )
                    break
            if not metric_found:
                raise ValueError(
                    f"{self.metric} metric not found. Please pass a string of the "
                    "name of a metric in scipy.spatial.distance or "
                    "distances.Distance, or pass a metric function directly. For a "
                    "list of available metrics, see: "
                    "https://sidchaini.github.io/DistClassiPy/distances.html or "
                    "https://docs.scipy.org/doc/scipy/reference/spatial.distance.html"
                )

    def fit(self, X: np.array, y: np.array, feat_labels: list[str] = None):
        """Calculate the feature space centroid for all classes.

        This function calculates the feature space centroid in the training
        set (X, y) for all classes using the central statistic. If scaling
        is enabled, it also calculates the appropriate dispersion statistic.
        This involves computing the centroid for every class in the feature space and
        optionally calculating the kernel density estimate and 1-dimensional distance.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.
        y : array-like of shape (n_samples,)
            The target values (class labels).
        feat_labels : list of str, optional, default=None
            The feature labels. If not provided, default labels representing feature
            number will be used.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        X, y = check_X_y(X, y)
        self.classes_ = unique_labels(y)
        self.n_features_in_ = X.shape[
            1
        ]  # Number of features seen during fit - required for sklearn compatibility.

        self.initialize_metric_function()

        if feat_labels is None:
            feat_labels = [f"Feature_{x}" for x in range(X.shape[1])]

        centroid_list = []
        for cur_class in self.classes_:
            cur_X = X[y == cur_class]
            if self.central_stat == "median":
                centroid_list.append(np.median(cur_X, axis=0).ravel())
            elif self.central_stat == "mean":
                centroid_list.append(np.mean(cur_X, axis=0).ravel())
        df_centroid = pd.DataFrame(
            data=np.array(centroid_list), index=self.classes_, columns=feat_labels
        )
        self.df_centroid_ = df_centroid

        if self.scale and self.dispersion_stat == "std":
            std_list = []
            for cur_class in self.classes_:
                cur_X = X[y == cur_class]
                # Note we're using ddof=1 because we're dealing with a sample.
                # See more: https://stackoverflow.com/a/46083501/10743245
                std_list.append(np.std(cur_X, axis=0, ddof=1).ravel())
            df_std = pd.DataFrame(
                data=np.array(std_list), index=self.classes_, columns=feat_labels
            )
            self.df_std_ = df_std
        elif self.scale and self.dispersion_stat == "iqr":

            iqr_list = []

            for cur_class in self.classes_:
                cur_X = X[y == cur_class]
                # Note we're using ddof=1 because we're dealing with a sample.
                # See more: https://stackoverflow.com/a/46083501/10743245
                iqr_list.append(
                    np.quantile(cur_X, q=0.75, axis=0).ravel()
                    - np.quantile(cur_X, q=0.25, axis=0).ravel()
                )
            df_iqr = pd.DataFrame(
                data=np.array(iqr_list), index=self.classes_, columns=feat_labels
            )
            self.df_iqr_ = df_iqr

        if self.calculate_kde:
            warnings.warn(
                "KDE calculation is deprecated and will be removed in version 0.2.0",
                DeprecationWarning,
            )
            self.kde_dict_ = {}

            for cl in self.classes_:
                subX = X[y == cl]
                # Implement the following in an if-else to save computational time.
                # kde = KernelDensity(bandwidth='scott', metric=self.metric)
                # kde.fit(subX)
                kde = KernelDensity(
                    bandwidth="scott",
                    metric="pyfunc",
                    metric_params={"func": self.metric_fn_},
                )
                kde.fit(subX)
                self.kde_dict_[cl] = kde
        self.is_fitted_ = True

        return self

    def predict(self, X: np.array):
        """Predict the class labels for the provided X.

        The prediction is based on the distance of each data point in the input sample
        to the centroid for each class in the feature space. The predicted class is the
        one whose centroid is the closest to the input sample.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray of shape (n_samples,)
            The predicted classes.
        """
        check_is_fitted(self, "is_fitted_")
        X = check_array(X)

        if not self.scale:
            dist_arr = scipy.spatial.distance.cdist(
                XA=X, XB=self.df_centroid_.to_numpy(), metric=self.metric_arg_
            )

        else:
            dist_arr_list = []

            if self.dispersion_stat == "std":
                # Clip to avoid zero error later
                wtdf = 1 / np.clip(self.df_std_, a_min=np.finfo(float).eps, a_max=None)
            elif self.dispersion_stat == "iqr":
                wtdf = 1 / np.clip(self.df_iqr_, a_min=np.finfo(float).eps, a_max=None)

            for cl in self.classes_:
                XB = self.df_centroid_.loc[cl].to_numpy().reshape(1, -1)
                w = wtdf.loc[cl].to_numpy()  # 1/std dev
                XB = XB * w  # w is for this class only
                XA = X * w  # w is for this class only
                cl_dist = scipy.spatial.distance.cdist(
                    XA=XA, XB=XB, metric=self.metric_arg_
                )
                dist_arr_list.append(cl_dist)
            dist_arr = np.column_stack(dist_arr_list)

        y_pred = self.classes_[dist_arr.argmin(axis=1)]
        return y_pred

    def predict_and_analyse(self, X: np.array):
        """Predict the class labels for the provided X and perform analysis.

        The prediction is based on the distance of each data point in the input sample
        to the centroid for each class in the feature space. The predicted class is the
        one whose centroid is the closest to the input sample.

        The analysis involves saving all calculated distances and confidences as an
        attribute for inspection and analysis later.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray of shape (n_samples,)
            The predicted classes.
        """
        check_is_fitted(self, "is_fitted_")
        X = check_array(X)

        if not self.scale:
            dist_arr = scipy.spatial.distance.cdist(
                XA=X, XB=self.df_centroid_.to_numpy(), metric=self.metric_arg_
            )

        else:
            dist_arr_list = []

            if self.dispersion_stat == "std":
                # Clip to avoid zero error later
                wtdf = 1 / np.clip(self.df_std_, a_min=np.finfo(float).eps, a_max=None)
            elif self.dispersion_stat == "iqr":
                wtdf = 1 / np.clip(self.df_iqr_, a_min=np.finfo(float).eps, a_max=None)

            for cl in self.classes_:
                XB = self.df_centroid_.loc[cl].to_numpy().reshape(1, -1)
                w = wtdf.loc[cl].to_numpy()  # 1/std dev
                XB = XB * w  # w is for this class only
                XA = X * w  # w is for this class only
                cl_dist = scipy.spatial.distance.cdist(
                    XA=XA, XB=XB, metric=self.metric_arg_
                )
                dist_arr_list.append(cl_dist)
            dist_arr = np.column_stack(dist_arr_list)

        self.centroid_dist_df_ = pd.DataFrame(
            data=dist_arr, index=np.arange(X.shape[0]), columns=self.classes_
        )
        self.centroid_dist_df_.columns = [
            f"{ind}_dist" for ind in self.centroid_dist_df_.columns
        ]

        y_pred = self.classes_[dist_arr.argmin(axis=1)]

        if self.calculate_kde:
            warnings.warn(
                "KDE calculation in predict_and_analyse is deprecated "
                "and will be removed in version 0.2.0",
                DeprecationWarning,
            )
            # NEW: Rescale in terms of median likelihoods - calculate here
            scale_factors = np.exp(
                [
                    self.kde_dict_[cl].score_samples(
                        self.df_centroid_.loc[cl].to_numpy().reshape(1, -1)
                    )[0]
                    for cl in self.classes_
                ]
            )

            likelihood_arr = []
            for k in self.kde_dict_.keys():
                log_pdf = self.kde_dict_[k].score_samples(X)
                likelihood_val = np.exp(log_pdf)
                likelihood_arr.append(likelihood_val)
            self.likelihood_arr_ = np.array(likelihood_arr).T

            # NEW: Rescale in terms of median likelihoods - rescale here
            self.likelihood_arr_ = self.likelihood_arr_ / scale_factors
        if self.calculate_1d_dist:
            warnings.warn(
                "calculate_1d_dist is deprecated and will be removed in version 0.2.0",
                DeprecationWarning,
            )
            conf_cl = []
            Xdf_temp = pd.DataFrame(data=X, columns=self.df_centroid_.columns)
            for cl in self.classes_:
                sum_1d_dists = np.zeros(shape=(len(Xdf_temp)))
                for feat in Xdf_temp.columns:
                    dists = scipy.spatial.distance.cdist(
                        XA=np.zeros(shape=(1, 1)),
                        XB=(self.df_centroid_.loc[cl] - Xdf_temp)[feat]
                        .to_numpy()
                        .reshape(-1, 1),
                        metric=self.metric_arg_,
                    ).ravel()
                    if self.scale and self.dispersion_stat == "std":
                        sum_1d_dists = sum_1d_dists + dists / self.df_std_.loc[cl, feat]
                    elif self.scale and self.dispersion_stat == "std":
                        sum_1d_dists = sum_1d_dists + dists / self.df_iqr_.loc[cl, feat]
                    else:
                        sum_1d_dists = sum_1d_dists + dists
                confs = 1 / np.clip(sum_1d_dists, a_min=np.finfo(float).eps, a_max=None)
                conf_cl.append(confs)
            conf_cl = np.array(conf_cl)
            self.conf_cl_ = conf_cl
        self.analyis_ = True

        return y_pred

    def calculate_confidence(self, method: str = "distance_inverse"):
        """Calculate the confidence for each prediction.

        The confidence is calculated based on either the distance of each data point to
        the centroids of the training data, optionally the kernel density estimate or
        1-dimensional distance.

        Parameters
        ----------
        method : {"distance_inverse", "1d_distance_inverse", "kde_likelihood"},
                 default="distance_inverse"
            The method to use for calculating confidence. Default is
            'distance_inverse'.
            .. deprecated:: 0.2.0
                The methods '1d_distance_inverse' and
                'kde_likelihood' will be removed in version 0.2.0.
        """
        check_is_fitted(self, "is_fitted_")
        if not hasattr(self, "analyis_"):
            raise ValueError(
                "Use predict_and_analyse() instead of predict() for "
                "confidence calculation."
            )

        # Calculate confidence for each prediction
        if method == "distance_inverse":
            self.confidence_df_ = 1 / np.clip(
                self.centroid_dist_df_, a_min=np.finfo(float).eps, a_max=None
            )
            self.confidence_df_.columns = [
                x.replace("_dist", "_conf") for x in self.confidence_df_.columns
            ]

        elif method == "1d_distance_inverse":
            warnings.warn(
                "The '1d_distance_inverse' method is deprecated "
                "and will be removed in version 0.2.0",
                DeprecationWarning,
            )
            if not self.calculate_1d_dist:
                raise ValueError(
                    "method='1d_distance_inverse' is only valid if calculate_1d_dist "
                    "is set to True"
                )
            self.confidence_df_ = pd.DataFrame(
                data=self.conf_cl_.T, columns=[f"{x}_conf" for x in self.classes_]
            )

        elif method == "kde_likelihood":
            warnings.warn(
                "The 'kde_likelihood' method is deprecated and will be "
                "removed in version 0.2.0",
                DeprecationWarning,
            )
            if not self.calculate_kde:
                raise ValueError(
                    "method='kde_likelihood' is only valid if calculate_kde is set "
                    "to True"
                )

            self.confidence_df_ = pd.DataFrame(
                data=self.likelihood_arr_,
                columns=[f"{x}_conf" for x in self.kde_dict_.keys()],
            )

        return self.confidence_df_.to_numpy()



from distclassipy.distances import Distance

from hypothesis import given, strategies as st

import numpy as np

import pytest

# Initialize the Distance class to use its methods for testing
distance = Distance()

# Strategy to generate arrays of floats
arrays = st.integers(min_value=1, max_value=20).flatmap(
    lambda n: st.tuples(
        st.lists(
            st.floats(
                min_value=1,
                max_value=10e3,
                allow_nan=False,
                allow_infinity=False,
                width=32,
            ),
            min_size=n,
            max_size=n,
        ).map(np.array),
        st.lists(
            st.floats(
                min_value=1,
                max_value=10e3,
                allow_nan=False,
                allow_infinity=False,
                width=32,
            ),
            min_size=n,
            max_size=n,
        ).map(np.array),
    )
)
# List of all distance metrics
_ALL_METRICS = [
    "euclidean",
    "braycurtis",
    "canberra",
    "cityblock",
    "chebyshev",
    "clark",
    "correlation",
    "cosine",
    "hellinger",
    "jaccard",
    "lorentzian",
    "marylandbridge",
    "meehl",
    "motyka",
    "soergel",
    "wave_hedges",
    "kulczynski",
    "add_chisq",
]


@pytest.mark.parametrize(
    "metric", [m for m in _ALL_METRICS if m not in ["marylandbridge"]]
)  # Note: Maryland bridge is excluded as it fails this test.
@given(arrays)
def test_non_negative(metric, data):
    u, v = data
    assert getattr(distance, metric)(u, v) >= 0


@pytest.mark.parametrize(
    "metric", [m for m in _ALL_METRICS if m not in ["motyka"]]
)  # Note: Motyka is excluded as it fails this test.
@given(arrays)
def test_self_distance(metric, data):
    u, _ = data
    assert getattr(distance, metric)(u, u) == 0


@pytest.mark.parametrize("metric", _ALL_METRICS)
@given(arrays)
def test_symmetry(metric, data):
    u, v = data
    assert getattr(distance, metric)(u, v) == getattr(distance, metric)(v, u)
